#!/usr/bin/perl -I/usr/sausalito/perl
#
# $Id: vz_vserver_backup.pl
#
# This script performs local and remote backing up of virtual private servers. 
# Only VPS's which have either local or remote backup enabled in the GUI will be backed up!
#

use CCE;
use POSIX qw(isalpha);
use Socket;
use Sys::Hostname;
use File::Spec;
use List::MoreUtils qw(uniq);
use POSIX qw(strftime);
use MIME::Lite;
use Net::SFTP::Foreign;
use File::Basename;
use Data::Dumper;

my $cce = new CCE;
$cce->connectuds();

$oid = $cce->event_oid();
$obj = $cce->event_object();

# Helper to redirect SSH errors to /dev/null:
open $ssherr, '>', '/dev/null'; # or die "Unable to open /dev/null";

$host = hostname();
$addr = inet_ntoa(scalar(gethostbyname($host)) || 'localhost');
$now = localtime time;
$datetime = strftime('%Y-%m-%d %H:%M:%S', localtime);

$DEBUG = "1";

# Get local / remote switch:
$type_switch = $ARGV[0];
if ($type_switch eq "local") {
    $b_type = "local";
    $B_type = "Local";
    $s_type = "local";
}
elsif ($type_switch eq "remote") {
    $b_type = "remote";
    $B_type = "Remote";
    $s_type = "remote";
}
else {
    $b_type = "local & remote";
    $B_type = "Local & Remote";
    $s_type = "both";
}

# Find out if there is anything to back up (we may be cluster secondary after all!)
if (!-d "/vz/private") {
    print STDERR "INFO: This node is currently secondary in a cluster.\n";
    print STDERR "So there is nothing to backup on this node.\n\n";
    exit 1;
}

# Get settings out of CODB to determine which type of backup is enabled:
@oids = $cce->find('VserverMasterNode');
if (!defined($oids[0])) {
    print STDERR "VZBACKUP: Sorry, no VserverMasterNode object in CCE found!\n";
    exit 1;
}

# Remove stale lockfiles if they're older than 180 minutes:  
if (-f "/tmp/vz_vserver_backup.lock") {
    system("find /tmp/vz_vserver_backup.lock -type f -cmin +180 -print | xargs rm >/dev/null 2>&1");
}

# Create lockfile:
if (-f "/tmp/vz_vserver_backup.lock") {
    print STDERR "ERROR: Found lockfile /tmp/vz_vserver_backup.lock!\n";
    print STDERR "Please check that no other instance of this program runs!\n\n";
    exit 1;
}

($ok, $solbackupValues) = $cce->get($oids[0]);
if ($ok) {
    if ($solbackupValues->{'rsync_enabled'} == "1") {
       $using_vzdump = "0";
        # Use RSYNC:
        &type_rsync;
    }
    else {
        # Do nothing:
        print STDERR "INFO: Backup is not configured on this node.\n";
        print STDERR "So there is nothing to do here. Please configure\n";
        print STDERR "the backup through the GUI first!\n";
    }
    # Remove lockfile:
    system("/bin/rm -f /tmp/vz_vserver_backup.lock");
}

$cce->bye('SUCCESS');
exit 0;

## New RSYNC type backup below:

sub type_rsync {

    # Email generation:
    open (EMAIL, '>>/tmp/.backup_mail');
    print EMAIL "=================================================================\n";
    print EMAIL "Aventurin{e} $B_type Backup on $host ($addr) \n";
    print EMAIL "Starting up at: $now\n";
    print EMAIL "=================================================================\n\n";

    # Get settings for RSYNC related backup:
    &rsync_settings;

    $vzlist = `/usr/sbin/vzlist -a -H -o vpsid,status,layout,ip,hostname | /bin/sed 's/^[ \t]*//;s/[ \t]*\$//'`;
    @vzlist_arr = split(/\n/, $vzlist);
    $VPSINFO = {};
    # Array of VPSIDs:
    @matrix_vpsids = ();
    $a = 0;
    foreach $vz_line (@vzlist_arr) {
        chomp($vz_line);
        next if $vz_line =~ /^\s*$/;               # skip blank lines
        next if $vz_line =~ /^VEID/;               # skip lines starting with VEID
        next if $vz_line =~ /^CTID/;               # skip lines starting with VEID
        my (@row) = split (/\s+/, $vz_line);
        $VPSINFO->{$row[0]} = {'status' => $row[1], 'layout' => $row[2], 'ip' => $row[3], 'hostname' => $row[4]};
        $a++;
    }

    # Handle local RSYNC backup if local backup is enabled and supposed to run now:
    if (($MN_LocalBackupEnabled == "1") && (($s_type eq "local") || ($s_type eq "both"))) {

        print STDERR "Starting local RSYNC backup of VPS's ... \n\n";
        print EMAIL "Starting local RSYNC backup of VPS's ... \n";
        print EMAIL "=========================================\n\n";

        # Find all OpenVZ VPS's where local backup is enabled:
        $nolocal = "0";
        @oids = $cce->find('Vserver_Backup', {'VserverBackupLocalEnabled' => "1", 'virttype' => 'openvz'});
        if ($#oids == -1) {
                $nolocal = "1";
        }
        else {
            # Get all VPS's with enabled local backup:
            undef @vps_local;
            undef @sorted_includes;
            foreach $line (@oids) {
                ($ok, $system) = $cce->get($line);
                $func = push(@vps_local, $system->{vpsid});
            }
            @sorted_includes = sort { $a <=> $b } @vps_local;
        }

        # Handle early exit if nothing to do:
        if ($nolocal == "1") {
            print STDERR "No VPS has local backup enabled. \n";
            print EMAIL "No VPS has local backup enabled. \n\n";
        }
        else {

            # Start the backup of OpenVZ containers:
            foreach $vpsid (@sorted_includes) {
                $local_target_dir_vps = $local_location_rsync . "/" . $vpsid;
                $vps_fqdn = $VPSINFO->{$vpsid}->{'hostname'};
                $vps_ip = $VPSINFO->{$vpsid}->{'ip'};
                print STDERR "Performing local backup of VPS $vpsid to $local_target_dir_vps\n";
                print EMAIL "Performing local backup of VPS $vpsid to $local_target_dir_vps\n";

                $vps_rundir_cfg = '/vz/root/' . $vpsid . '/etc';
                if (($VPSINFO->{$vpsid}->{'status'} eq 'running') && (-d $vps_rundir_cfg)) {
                    # If need be, make config storage dir:
                    $vps_cfg_storage_dir = $vps_rundir_cfg . '/vzdump';
                    if (!-d $vps_cfg_storage_dir) {
                        system("mkdir -p $vps_cfg_storage_dir");
                    }
                    # Copy VPS config file:
                    system("cp /etc/vz/conf/$vpsid.conf $vps_cfg_storage_dir/vps.conf");
                }
                if (!-d "$local_target_dir_vps") {
                    system("mkdir -p $local_target_dir_vps");
                }

                # Simfs backup:
                if ($VPSINFO->{$vpsid}->{'layout'} eq 'simfs') {
                    print "VPS $vpsid is using Simfs. \n";
                    print EMAIL "VPS $vpsid is using Simfs. \n";
                    system("/usr/bin/rsync -q -ar /vz/private/$vpsid $local_location_rsync --delete-after");
                }
                else {
                    print "VPS $vpsid is using Ploop. \n";
                    print EMAIL "VPS $vpsid is using Ploop. \n";
                    system("rm -Rf $local_target_dir_vps");
                    system("mkdir -p $local_target_dir_vps");
                    if (-d $local_target_dir_vps) {
                        system("/bin/rm -Rf $local_target_dir_vps");
                        system("mkdir $local_target_dir_vps");
                    }
                    $temp_vps_dump = "/vz/vztmp/" . $vpsid;
                    system("rm -Rf $temp_vps_dump");
                    system("mkdir -p $temp_vps_dump");
                    system("/usr/sausalito/sbin/vzpbackup.sh $vpsid --suspend no --compact --backup-dir=$local_target_dir_vps --work-dir=$temp_vps_dump >/dev/null 2>&1 || :");
                    system("cp /etc/vz/conf/$vpsid.conf $local_location_rsync/$vpsid/vps.conf");
                    system("rm -Rf $temp_vps_dump")
                }

                # As local backup has just ended, we now update CODB with the info that says so:
                @vps_oid = $cce->find('Backup', {'vpsid' => $vpsid, 'origin' => 'rsync', 'location' => 'local', 'virttype' => 'openvz'});
                if ($#vps_oid < 0) {
                    # No CODB object for this type of backup yet. Creating one:
                    ($ok) = $cce->create('Backup', {
                        'origin' => 'rsync',
                        'vpsid' => $vpsid,
                        'virttype' => 'openvz',
                        'layout' => $VPSINFO->{$vpsid}->{'layout'},
                        'location' => 'local',
                        'LBtimestamp' => $datetime,
                        'storage' => $local_target_dir_vps,
                        'ip' => $vps_ip,
                        'hostname' => $vps_fqdn,
                        'timestamp' => time()
                    });
                }
                else {
                    # Object already present - updating it:
                    ($ok) = $cce->set($vps_oid[0], '', {
                        'origin' => 'rsync',
                        'vpsid' => $vpsid,
                        'virttype' => 'openvz',
                        'layout' => $VPSINFO->{$vpsid}->{'layout'},
                        'location' => 'local',
                        'LBtimestamp' => $datetime,
                        'storage' => $local_target_dir_vps,
                        'ip' => $vps_ip,
                        'hostname' => $vps_fqdn,
                        'timestamp' => time()
                    });
                }
                print STDERR "Finished local backup of VPS $vpsid \n\n";
                print EMAIL "Finished local backup of VPS $vpsid \n\n";
            }
        }
    }

    # Handle remote RSYNC backup if remote backup is enabled and supposed to run now:
    if (($MN_RemoteBackupEnabled == "1") && (($s_type eq "remote") || ($s_type eq "both"))) {

        print STDERR "Starting remote RSYNC backup of VPS's ... \n\n";
        print EMAIL "Starting remote RSYNC backup of VPS's ... \n";
        print EMAIL "=========================================\n\n";

        # Find all OpenVZ VPS's where remote backup is enabled:
        $noremote = "0";
        @oids = $cce->find('Vserver_Backup', {'VserverBackupRemoteEnabled' => "1", 'virttype' => 'openvz'});
        if ($#oids == -1) {
                $noremote = "1";
        }
        else {
            # Get all VPS's with enabled remote backup:
            undef @vps_remote;
            undef @sorted_includes;
            foreach $line (@oids) {
                ($ok, $system) = $cce->get($line);
                $vps_config = "/etc/vz/conf/" . $system->{vpsid} . ".conf";
                if (-f $vps_config) {
                    # VPS exists, push it into array for backing up:
                    $func = push(@vps_remote, $system->{vpsid});
                }
                else {
                    # VPS does not exists. This must be a backup entry from a removed VPS.
                    # Remove the 'Vserver_Backup' entry as it is no longer valid:
                    ($ok) = $cce->destroy($line);
                }
            }
            @sorted_includes = sort { $a <=> $b } @vps_remote;
        }

        # Handle early exit if nothing to do:
        if ($noremote == "1") {
            print STDERR "No VPS has remote backup enabled.\n";
            print EMAIL "No VPS has remote backup enabled. \n\n";
        }
        else {
            # Start the backup of OpenVZ containers:
            foreach $vpsid (@sorted_includes) {
                $remote_target_dir_vps = $remote_location_rsync . "/" . $vpsid;
                $vps_fqdn = $VPSINFO->{$vpsid}->{'hostname'};
                $vps_ip = $VPSINFO->{$vpsid}->{'ip'};
                print STDERR "Performing remote backup of VPS $vpsid to $remote_target_dir_vps\n";

                $vps_rundir_cfg = '/vz/root/' . $vpsid . '/etc';
                if (($VPSINFO->{$vpsid}->{'status'} eq 'running') && (-d $vps_rundir_cfg)) {
                    # If need be, make config storage dir:
                    $vps_cfg_storage_dir = $vps_rundir_cfg . '/vzdump';
                    if (!-d $vps_cfg_storage_dir) {
                        system("mkdir -p $vps_cfg_storage_dir");
                    }
                    # Copy VPS config file:
                    system("cp /etc/vz/conf/$vpsid.conf $vps_cfg_storage_dir/vps.conf");
                }

                # Handle --bwlimit:
                if (($solbackupValues->{'RsyncSpeedLimit'} ne "0") && ($solbackupValues->{'RsyncSpeedLimit'} ne "")) {
                    $RsyncSpeedLimit = $solbackupValues->{'RsyncSpeedLimit'};
                    $bwlimit = "--bwlimit=$RsyncSpeedLimit";
                }
                else {
                    $bwlimit = "";
                }

                # FQDN or IP:
                if ($targethost) {
                    $target = $targethost;
                }
                else {
                    $target = $targetip;
                }
                print EMAIL "Performing remote backup of VPS $vpsid to $target:$remote_target_dir_vps\n";

                # Simfs backup:
                if ($VPSINFO->{$vpsid}->{'layout'} eq 'simfs') {
                    print "VPS $vpsid is using Simfs. \n";
                    print EMAIL "VPS $vpsid is using Simfs. \n";
                    system("/usr/bin/rsync -q -ar $bwlimit -e 'ssh -p $port' /vz/private/$vpsid $target:$remote_location_rsync/ --delete-after");
                }
                else {
                    print "VPS $vpsid is using Ploop. \n";
                    print EMAIL "VPS $vpsid is using Ploop. \n";
                    $temp_vps_dump = "/vz/vztmp/" . $vpsid;
                    system("rm -Rf $temp_vps_dump");
                    system("mkdir -p $temp_vps_dump");
                    $temp_vps_local_storage = "/vz/vztmp/" . $vpsid . "-out";
                    system("rm -Rf $temp_vps_local_storage");
                    system("mkdir -p $temp_vps_local_storage");
                    system("/usr/sausalito/sbin/vzpbackup.sh $vpsid --suspend no --compact --backup-dir=$temp_vps_local_storage --work-dir=$temp_vps_dump >/dev/null 2>&1 || :");
                    system("cp /etc/vz/conf/$vpsid.conf $temp_vps_local_storage/vps.conf");
                    system("rm -Rf $temp_vps_dump");
                    system("mv $temp_vps_local_storage $temp_vps_dump");
                    system("/usr/bin/rsync -q -ar $bwlimit -e 'ssh -p $port' $temp_vps_dump $target:$remote_location_rsync/ --delete-after");
                    system("rm -Rf $temp_vps_dump");
                }

                # As remote backup has just ended, we now update CODB with the info that says so:
                @vps_oid = $cce->find('Backup', {'vpsid' => $vpsid, 'origin' => 'rsync', 'location' => 'remote', 'virttype' => 'openvz'});
                if ($#vps_oid < 0) {
                    # No CODB object for this type of backup yet. Creating one:
                    ($ok) = $cce->create('Backup', {
                        'origin' => 'rsync',
                        'vpsid' => $vpsid,
                        'virttype' => 'openvz',
                        'layout' => $VPSINFO->{$vpsid}->{'layout'},
                        'location' => 'remote',
                        'LBtimestamp' => $datetime,
                        'storage' => $remote_target_dir_vps,
                        'ip' => $vps_ip,
                        'hostname' => $vps_fqdn,
                        'timestamp' => time()
                    });
                }
                else {
                    # Object already present - updating it:
                    ($ok) = $cce->set($vps_oid[0], '', {
                        'origin' => 'rsync',
                        'vpsid' => $vpsid,
                        'virttype' => 'openvz',
                        'layout' => $VPSINFO->{$vpsid}->{'layout'},
                        'location' => 'remote',
                        'LBtimestamp' => $datetime,
                        'storage' => $remote_target_dir_vps,
                        'ip' => $vps_ip,
                        'hostname' => $vps_fqdn,
                        'timestamp' => time()
                    });
                }
                print STDERR "Finished remote backup of VPS $vpsid \n\n";
                print EMAIL "Finished remote backup of VPS $vpsid \n\n";
            }
        }
    }

    $now = localtime time;
    print EMAIL "=================================================================\n";
    print EMAIL "Aventurin{e} $B_type Backup on $host ($addr) \n";
    print EMAIL "Backup finished at: $now\n";
    print EMAIL "=================================================================\n\n";
    close (EMAIL);

    # Run extender:
    if ( -f "/usr/sausalito/sbin/vz_vserver_backup_extend.pl") {
        system("/usr/sausalito/sbin/vz_vserver_backup_extend.pl");
    }

    # Send the email:
    open (EMAIL, '/tmp/.backup_mail');
    while (<EMAIL>) {
        $body .= $_;
    }
    close (EMAIL);
    system("/bin/rm -f /tmp/.backup_mail");

    if ($body) {
        $body = $body_head . $body;
        $subject = $host . ": rsync backup finished";

        # Build the message using MIME::Lite instead:
        my $send_msg = MIME::Lite->new(
                From     => "root <root>",
                To       => "root <root>",
                Subject  => $subject,
                Data     => $body
        );   
            
        # Set content type:
        $send_msg->attr("content-type"         => "text/plain");
        $send_msg->attr("content-type.charset" => "ISO-8859-1"); 
         
        # Out with the email:
        $send_msg->send;
    }
}

# General Settings for RSYNC related backup:
sub rsync_settings {

    # Define values:
    $vzdump_enabled             = $solbackupValues->{'vzdump_enabled'};
    $rsync_enabled              = $solbackupValues->{'rsync_enabled'};
    $MN_RemoteBackupEnabled     = $solbackupValues->{'MN_RemoteBackupEnabled'};
    $MN_LocalBackupEnabled      = $solbackupValues->{'MN_LocalBackupEnabled'};
    $MN_BackupLocalLocation     = $solbackupValues->{'MN_BackupLocalLocation'};
    $targethost                 = $solbackupValues->{'MN_BackupRemoteTargetHost'};
    $targetip                   = $solbackupValues->{'MN_BackupRemoteTargetIP'};
    $path                       = $solbackupValues->{'MN_BackupRemoteLocation'};
    $user                       = $solbackupValues->{'MN_BackupRemoteTargetUser'};
    $pass                       = $solbackupValues->{'MN_BackupRemoteTargetUserPass'};
    $port                       = $solbackupValues->{'MN_BackupRemoteTargetPort'};
    $limit                      = $solbackupValues->{'RsyncSpeedLimit'};
    $my_maxfiles                = $solbackupValues->{'vzdump_maxfiles'};
    $kvm_support                = $solbackupValues->{'kvm_support'};

    # Assemble paths:  
    $local_location_vzdump = $solbackupValues->{'MN_BackupLocalLocation'} . "/vzbackup";
    $local_location_vzdump =~ s|//|/|g;
    $local_location_vzdump =~ s|/$||g;

    $remote_location_vzdump = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_vzdump =~ s|//|/|g;
    $remote_location_vzdump =~ s|/$||g;

    $local_location_rsync = $solbackupValues->{'MN_BackupLocalLocation'} . "/rsync";
    $local_location_rsync =~ s|//|/|g;  
    $local_location_rsync =~ s|/$||g;

    $remote_location_rsync = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_rsync =~ s|//|/|g;
    $remote_location_rsync =~ s|/$||g;

    # Local directory setup:
    if (!-d $local_location_rsync) {
        system("mkdir -p $local_location_rsync");
    }
    if (!-w $local_location_rsync) {
        print STDERR "Specified backup location $solbackupValues->{'MN_BackupLocalLocation'} is not writeable!\n";
        print STDERR "Defaulting backup location to /vz/backup \n\n";
        $local_location_rsync = "/vz/backup/rsync";
    }
}

$cce->bye('SUCCESS');
exit 0;

# 
# Copyright (c) 2015 Michael Stauber, SOLARSPEED.NET
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#     notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in 
#     the documentation and/or other materials provided with the 
#     distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#     contributors may be used to endorse or promote products derived 
#     from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 